from flow_prompt.responses import AIResponse
from dataclasses import dataclass
from openai.types.chat import ChatCompletionMessage as Message


@dataclass(kw_only=True)
class GeminiAIResponse(AIResponse):
    message: Message = None

    def get_message_str(self) -> str:
        return self.message.model_dump_json(indent=2)

    def __str__(self) -> str:
        result = (
            f"finish_reason: {self.finish_reason}\n"
            f"message: {self.get_message_str()}\n"
        )
        return result



import logging
import typing as t
from dataclasses import dataclass
from decimal import Decimal

from flow_prompt import Secrets, settings
from flow_prompt.ai_models.ai_model import AI_MODELS_PROVIDER
from flow_prompt.ai_models.attempt_to_call import AttemptToCall
from flow_prompt.ai_models.behaviour import AIModelsBehaviour, PromptAttempts
from flow_prompt.exceptions import (
    FlowPromptIsnotFoundError,
    RetryableCustomError
)
from flow_prompt.services.SaveWorker import SaveWorker
from flow_prompt.prompt.pipe_prompt import PipePrompt
from flow_prompt.prompt.user_prompt import UserPrompt
from flow_prompt.responses import AIResponse
from flow_prompt.services.flow_prompt import FlowPromptService
from flow_prompt.utils import current_timestamp_ms


logger = logging.getLogger(__name__)

@dataclass
class FlowPrompt:
    api_token: str = None
    openai_key: str = None
    openai_org: str = None
    claude_key: str = None
    gemini_key: str = None
    azure_keys: t.Dict[str, str] = None
    secrets: Secrets = None

    clients = {}

    def __post_init__(self):
        self.secrets = Secrets()
        if not self.azure_keys:
            if self.secrets.azure_keys:
                logger.debug(f"Using Azure keys from secrets")
                self.azure_keys = self.secrets.azure_keys
            else:
                logger.debug(f"Azure keys not found in secrets")
        if not self.api_token and self.secrets.API_TOKEN:
            logger.debug(f"Using API token from secrets")
            self.api_token = self.secrets.API_TOKEN
        if not self.openai_key and self.secrets.OPENAI_API_KEY:
            logger.debug(f"Using OpenAI API key from secrets")
            self.openai_key = self.secrets.OPENAI_API_KEY
        if not self.openai_org and self.secrets.OPENAI_ORG:
            logger.debug(f"Using OpenAI organization from secrets")
            self.openai_org = self.secrets.OPENAI_ORG
        if not self.gemini_key and self.secrets.GEMINI_API_KEY:
            logger.debug(f"Using Gemini API key from secrets")
            self.gemini_key = self.secrets.GEMINI_API_KEY
        if not self.claude_key and self.secrets.CLAUDE_API_KEY:
            logger.debug(f"Using Claude API key from secrets")
            self.claude_key = self.secrets.CLAUDE_API_KEY
        self.service = FlowPromptService()
        if self.openai_key:
            self.clients[AI_MODELS_PROVIDER.OPENAI] = {
                'organization': self.openai_org,
                'api_key': self.openai_key,
            }
        if self.azure_keys:
            if not self.clients.get(AI_MODELS_PROVIDER.AZURE):
                self.clients[AI_MODELS_PROVIDER.AZURE] = {}
            for realm, key_data in self.azure_keys.items():
                self.clients[AI_MODELS_PROVIDER.AZURE][realm] = {
                    'api_version': key_data.get("api_version", "2023-07-01-preview"),
                    'azure_endpoint': key_data["url"],
                    'api_key': key_data["key"],
                }
                logger.debug(f"Initialized Azure client for {realm} {key_data['url']}")
        if self.claude_key:
            self.clients[AI_MODELS_PROVIDER.CLAUDE] = {'api_key': self.claude_key}
        if self.gemini_key:
            self.clients[AI_MODELS_PROVIDER.GEMINI] = {'api_key': self.gemini_key}
        self.worker = SaveWorker()


    def call(
        self,
        prompt_id: str,
        context: t.Dict[str, str],
        behaviour: AIModelsBehaviour,
        params: t.Dict[str, t.Any] = {},
        version: str = None,
        count_of_retries: int = None,
        test_data: dict = {},
        stream_function: t.Callable = None,
        check_connection: t.Callable = None,
        stream_params: dict = {},
    ) -> AIResponse:
        
        """
        Call flow prompt with context and behaviour
        """
        
        logger.debug(f"Calling {prompt_id}")
        start_time = current_timestamp_ms()
        pipe_prompt = self.get_pipe_prompt(prompt_id, version)
        prompt_attempts = PromptAttempts(behaviour, count_of_retries=count_of_retries)

        while prompt_attempts.initialize_attempt():
            current_attempt = prompt_attempts.current_attempt
            user_prompt = pipe_prompt.create_prompt(current_attempt)
            calling_messages = user_prompt.resolve(context)
            
            """
            Create CI/CD when calling first time
            """
            try:
                result = current_attempt.ai_model.call(
                    calling_messages.get_messages(),
                    calling_messages.max_sample_budget,
                    stream_function=stream_function,
                    check_connection=check_connection,
                    stream_params=stream_params,
                    client_secrets=self.clients[current_attempt.ai_model.provider],
                    **params,
                )

                sample_budget = self.calculate_budget_for_text(
                    user_prompt, result.get_message_str()
                )
                result.metrics.price_of_call = self.get_price(
                    current_attempt,
                    sample_budget,
                    calling_messages.prompt_budget,
                )
                result.metrics.sample_tokens_used = sample_budget
                result.metrics.prompt_tokens_used = calling_messages.prompt_budget
                result.metrics.ai_model_details = (
                    current_attempt.ai_model.get_metrics_data()
                )
                result.metrics.latency = current_timestamp_ms() - start_time

                if settings.USE_API_SERVICE and self.api_token:
                    self.worker.add_task(
                        self.api_token,
                        pipe_prompt.service_dump(),
                        context,
                        result,
                        test_data
                    )
                return result
            except RetryableCustomError as e:
                logger.error(
                    f"Attempt failed: {prompt_attempts.current_attempt} with retryable error: {e}"
                )
            except Exception as e:
                logger.exception(
                    f"Attempt failed: {prompt_attempts.current_attempt} with non-retryable error: {e}"
                )
                raise e

    def get_pipe_prompt(self, prompt_id: str, version: str = None) -> PipePrompt:
        """
        if the user has keys:  lib -> service: get_actual_prompt(local_prompt) -> Service:
        generates hash of the prompt;
        check in Redis if that record is the latest; if yes -> return 200, else
        checks if that record exists with that hash;
        if record exists and it's not the last - then we load the latest published prompt; - > return  200 + the last record
        add a new record in storage, and adding that it's the latest published prompt; -> return 200
        update redis with latest record;
        """
        logger.debug(f"Getting pipe prompt {prompt_id}")
        if (
            settings.USE_API_SERVICE
            and self.api_token
            and settings.RECEIVE_PROMPT_FROM_SERVER
        ):
            prompt_data = None
            prompt = settings.PIPE_PROMPTS.get(prompt_id)
            if prompt:
                prompt_data = prompt.service_dump()
            try:
                response = self.service.get_actual_prompt(
                    self.api_token, prompt_id, prompt_data, version
                )
                if not response.is_taken_globally:
                    prompt.version = response.version
                    return prompt
                response.prompt["version"] = response.version
                return PipePrompt.service_load(response.prompt)
            except Exception as e:
                logger.exception(f"Error while getting prompt {prompt_id}: {e}")
                if prompt:
                    return prompt
                else:
                    logger.exception(f"Prompt {prompt_id} not found")
                    raise FlowPromptIsnotFoundError()

        else:
            return settings.PIPE_PROMPTS[prompt_id]

    def calculate_budget_for_text(self, user_prompt: UserPrompt, text: str) -> int:
        if not text:
            return 0
        return len(user_prompt.encoding.encode(text))

    def get_price(
        self, attempt: AttemptToCall, sample_budget: int, prompt_budget: int
    ) -> Decimal:
        return attempt.ai_model.get_prompt_price(prompt_budget) + attempt.ai_model.get_sample_price(prompt_budget, sample_budget)


from flow_prompt.ai_models.ai_model import AI_MODELS_PROVIDER, AIModel
import logging

from flow_prompt.ai_models.constants import C_1M, C_128K
from flow_prompt.responses import AIResponse
from decimal import Decimal
from enum import Enum

import typing as t
from dataclasses import dataclass

from flow_prompt.ai_models.gemini.responses import GeminiAIResponse

from flow_prompt.ai_models.utils import get_common_args
from openai.types.chat import ChatCompletionMessage as Message
from flow_prompt.responses import Prompt
from flow_prompt.exceptions import RetryableCustomError, ConnectionLostError
import google.generativeai as genai


logger = logging.getLogger(__name__)


FLASH = "gemini-1.5-flash"
PRO = "gemini-1.5-pro"
PRO_1_0 = "gemini-1.0-pro"



class FamilyModel(Enum):
    flash = "Gemini 1.5 Flash"
    pro = "Gemini 1.5 Pro"
    pro_1_0 = "Gemini 1.0 Pro"


DEFAULT_PRICING = {
    "price_per_prompt_1k_tokens": Decimal(0.00035),
    "price_per_sample_1k_tokens": Decimal(0.00105),
}

GEMINI_AI_PRICING = {
    FamilyModel.flash.value: {
        C_128K: {
            "price_per_prompt_1k_tokens": Decimal(0.0075),
            "price_per_sample_1k_tokens": Decimal(0.030),
        },
        C_1M: {
            "price_per_prompt_1k_tokens": Decimal(0.015),
            "price_per_sample_1k_tokens": Decimal(0.060),
        }
    },
    FamilyModel.pro_1_0.value: {
        C_1M: {
            "price_per_prompt_1k_tokens": Decimal(0.0005),
            "price_per_sample_1k_tokens": Decimal(0.0015),
        }
    },
    FamilyModel.pro.value: {
        C_128K: {
            "price_per_prompt_1k_tokens": Decimal(0.0035),
            "price_per_sample_1k_tokens": Decimal(0.0105),
        },
        C_1M: {
            "price_per_prompt_1k_tokens": Decimal(0.007),
            "price_per_sample_1k_tokens": Decimal(0.021),
        }
    },
}


@dataclass(kw_only=True)
class GeminiAIModel(AIModel):
    model: str
    max_tokens: int = C_1M
    gemini_model: genai.GenerativeModel = None
    provider: AI_MODELS_PROVIDER = AI_MODELS_PROVIDER.GEMINI
    family: str = None

    def __post_init__(self):
        self.model = self.model.lower()
        if FLASH in self.model:
            self.family = FamilyModel.flash.value
        elif PRO_1_0 in self.model:
            self.family = FamilyModel.pro_1_0.value
        elif PRO in self.model:
            self.family = FamilyModel.pro.value
        else:
            logger.warning(
                f"Unknown family for {self.model}. Please add it obviously. Setting as Gemini 1.5 Flash"
            )
            self.family = FamilyModel.flash.value

    def call(self, messages: t.List[dict], max_tokens: int, client_secrets: dict = {}, **kwargs) -> AIResponse:
        genai.configure(api_key=client_secrets["api_key"])
        self.gemini_model = genai.GenerativeModel(self.model)
        common_args = get_common_args(max_tokens)
        kwargs = {
            **{
                "messages": messages,
            },
            **common_args,
            **client_secrets,
            **client_secrets,
            **self.get_params(),
            **kwargs,
        }

        logger.debug(
            f"Calling {messages} with max_tokens {max_tokens} and kwargs {kwargs}"
        )

        stream_function = kwargs.get("stream_function")
        check_connection = kwargs.get("check_connection")
        stream_params = kwargs.get("stream_params")

        # Parse only prompt content due to gemini call specifics
        prompt = '\n\n'.join([obj["content"] for obj in messages])

        content = ""

        try:
            if not kwargs.get('stream'):
                response = self.gemini_model.generate_content(prompt, stream=False)
                content = response.text
            else:
                response = self.gemini_model.generate_content(prompt, stream=True)
                idx = 0
                for chunk in response:
                    if idx % 5 == 0:
                        idx = 0
                        if not check_connection(**stream_params):
                            raise ConnectionLostError("Connection was lost!")
                    stream_function(chunk.text, **stream_params)
                    content += chunk.text
                    idx += 1

            return GeminiAIResponse(
                message=Message(content=content, role="assistant"),
                content=content,
                prompt=Prompt(
                    messages=kwargs.get("messages"),
                    functions=kwargs.get("tools"),
                    max_tokens=max_tokens,
                    temperature=kwargs.get("temperature"),
                    top_p=kwargs.get("top_p"),
                ),
            )

        except Exception as e:
            logger.exception("[GEMINIAI] failed to handle chat stream", exc_info=e)
            raise RetryableCustomError(f"Gemini AI call failed!")

    def name(self) -> str:
        return self.model

    def get_params(self) -> t.Dict[str, t.Any]:
        return {
            "model": self.model,
            "max_tokens": self.max_tokens,
        }

    def get_metrics_data(self) -> t.Dict[str, t.Any]:
        return {
            "model": self.model,
            "max_tokens": self.max_tokens,
        }
    

    def get_prompt_price(self, count_tokens: int) -> Decimal:
        for key in sorted(GEMINI_AI_PRICING[self.family].keys()):
            if count_tokens < key:
                logger.info(f"Prompt price for {count_tokens} tokens is {GEMINI_AI_PRICING[self.family][key]['price_per_prompt_1k_tokens'] * Decimal(count_tokens) / 1000}")
                return self._decimal(GEMINI_AI_PRICING[self.family][key]["price_per_prompt_1k_tokens"] * Decimal(count_tokens) / 1000)
        
        return self._decimal(self.price_per_prompt_1k_tokens * Decimal(count_tokens) / 1000)
    
    def get_sample_price(self, prompt_sample, count_tokens: int) -> Decimal:
        for key in sorted(GEMINI_AI_PRICING[self.family].keys()):
            if prompt_sample < key:
                logger.info(f"Sample price for {count_tokens} tokens is {GEMINI_AI_PRICING[self.family][key]['price_per_prompt_1k_tokens'] * Decimal(count_tokens) / 1000}")
                return self._decimal(GEMINI_AI_PRICING[self.family][key]["price_per_sample_1k_tokens"] * Decimal(count_tokens) / 1000)
        return self._decimal(self.price_per_sample_1k_tokens * Decimal(count_tokens) / 1000)


from dataclasses import dataclass
import json
import logging

import yaml

from flow_prompt.exceptions import NotParsedResponseException
from flow_prompt.responses import AIResponse

logger = logging.getLogger(__name__)


@dataclass
class Tag:
    start_tag: str
    end_tag: str
    include_tag: bool
    is_right_find_end_ind: bool = False


@dataclass
class TaggedContent:
    content: str
    start_ind: int
    end_ind: int
    parsed_content: any = None


def get_yaml_from_response(response: AIResponse):
    content, start_ind, end_ind = _get_format_from_response(
        response, [Tag("```yaml", "```", 0, 0), Tag("```", "```", 0, 0)]
    )
    parsed_content = None
    if content:
        try:
            parsed_content = yaml.safe_load(content)
        except Exception as e:
            logger.exception(f"Couldn't parse yaml:\n{content}")
        return TaggedContent(
            content=content,
            parsed_content=parsed_content,
            start_ind=start_ind,
            end_ind=end_ind,
        )


def get_json_from_response(response: AIResponse, start_from: int = 0) -> TaggedContent:
    content, start_ind, end_ind = _get_format_from_response(
        response,
        [Tag("```json", "\n```", 0), Tag("```json", "```", 0), Tag("{", "}", 1)],
        start_from=start_from,
    )
    if content:
        try:
            json_response = eval(content)
            return TaggedContent(
                content=content,
                parsed_content=json_response,
                start_ind=start_ind,
                end_ind=end_ind,
            )
        except Exception as e:
            try:
                json_response = json.loads(content)
                return TaggedContent(
                    content=content,
                    parsed_content=json_response,
                    start_ind=start_ind,
                    end_ind=end_ind,
                )
            except Exception as e:
                logger.exception(f"Couldn't parse json:\n{content}")
                raise NotParsedResponseException()


def _get_format_from_response(
    response: AIResponse, tags: list[Tag], start_from: int = 0
):
    start_ind, end_ind = 0, -1
    content = response.response[start_from:]
    for t in tags:
        start_ind = content.find(t.start_tag)
        if t.is_right_find_end_ind:
            end_ind = content.rfind(t.end_tag, start_ind + len(t.start_tag))
        else:
            end_ind = content.find(t.end_tag, start_ind + len(t.start_tag))
        if start_ind != -1:
            try:
                if t.include_tag:
                    end_ind += len(t.end_tag)
                else:
                    start_ind += len(t.start_tag)
                response_tagged = content[start_ind:end_ind].strip()
                return response_tagged, start_from + start_ind, start_from + end_ind
            except Exception as e:
                logger.exception(f"Couldn't parse json:\n{content}")
    return None, 0, -1
